import pytest


@pytest.fixture
def reset_helpers_namespace(request):
    try:
        yield
    finally:
        plugin = request.config.pluginmanager.get_plugin("helpers-namespace")
        plugin._registry.clear()


def test_namespace(testdir):
    testdir.makeconftest(
        """
        import pytest

        @pytest.helpers.register
        def foo(bar):
            return bar
        """
    )

    testdir.makepyfile(
        """
        import pytest

        def test_helpers():
            assert pytest.helpers.foo(True) is True
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines(
        [
            "test_namespace.py PASSED",
        ]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret == 0


def test_nested_namespace(testdir):
    testdir.makeconftest(
        """
        import pytest

        @pytest.helpers.foo.bar.register
        def foo(bar):
            return bar
        """
    )

    testdir.makepyfile(
        """
        import pytest

        def test_helpers():
            assert pytest.helpers.foo.bar.foo(True) is True
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines(
        [
            "test_nested_namespace.py PASSED",
        ]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret == 0


def test_unregistered_namespace(testdir):
    testdir.makepyfile(
        """
        import pytest

        def test_helpers():
            with pytest.raises(RuntimeError) as exc:
                assert pytest.helpers.foo(True) is True
            assert 'The helper being called was not registred' in str(exc)
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines(
        [
            "test_unregistered_namespace.py PASSED",
        ]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret == 0


def test_namespace_override(testdir):
    testdir.makeconftest(
        """
        import pytest

        @pytest.helpers.foo.register
        def bar(bar):
            return bar

        @pytest.helpers.register
        def foo(bar):
            return bar
        """
    )
    testdir.makepyfile(
        """
        import pytest

        def test_helpers():
            with pytest.raises(RuntimeError) as exc:
                assert pytest.helpers.foo(True) is True
            assert 'The helper being called was not registred' in str(exc)
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stderr.fnmatch_lines(
        ["*RuntimeError: A helper function is already registered under the name: foo"]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret != 0


def test_namespace_override_2(testdir):
    testdir.makeconftest(
        """
        import pytest

        @pytest.helpers.register
        def foo(bar):
            return bar

        @pytest.helpers.foo.register
        def bar(bar):
            return bar
        """
    )
    testdir.makepyfile(
        """
        import pytest

        def test_helpers():
            with pytest.raises(RuntimeError) as exc:
                assert pytest.helpers.foo(True) is True
            assert 'The helper being called was not registred' in str(exc)
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stderr.fnmatch_lines(
        ["*RuntimeError: A namespace is already registered under the name: bar"]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret != 0


def test_helper_override(testdir):
    testdir.makeconftest(
        """
        import pytest

        @pytest.helpers.register
        def foo(bar):
            return bar

        @pytest.helpers.register
        def foo(bar):
            return bar
        """
    )
    testdir.makepyfile(
        """
        import pytest

        def test_helpers():
            with pytest.raises(RuntimeError) as exc:
                assert pytest.helpers.foo(True) is True
            assert 'The helper being called was not registred' in str(exc)
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stderr.fnmatch_lines(
        ["*RuntimeError: A helper function is already registered under the name: foo"]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret != 0


def test_helper_as_regular_function(testdir):
    testdir.makepyfile(
        """
        import pytest

        @pytest.helpers.register
        def foo():
            return 'bar'

        def test_helpers():
            assert pytest.helpers.foo() == 'bar'
            assert foo() == 'bar'
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines(
        [
            "test_helper_as_regular_function.py PASSED",
        ]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret == 0


def test_helper_with_custom_name(testdir):
    testdir.makepyfile(
        """
        import pytest

        @pytest.helpers.register('jump')
        def foo():
            return 'bar'

        def test_helpers():
            assert pytest.helpers.jump() == 'bar'
            assert foo() == 'bar'
            print('PASSED')
    """
    )

    result = testdir.runpytest_subprocess("-s")

    # fnmatch_lines does an assertion internally
    result.stdout.fnmatch_lines(
        [
            "test_helper_with_custom_name.py PASSED",
        ]
    )

    # make sure that that we get a '0' exit code for the test suite
    assert result.ret == 0


@pytest.mark.usefixtures("reset_helpers_namespace")
def test_helper_contains_method():
    assert "bar" not in pytest.helpers

    @pytest.helpers.register
    def bar():
        return True

    assert "bar" in pytest.helpers
    assert pytest.helpers.bar() is True



"""
pytest_helpers_namespace.plugin
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Pytest Helpers Namespace Plugin
"""
from functools import partial
from functools import wraps

import pytest

try:  # pragma: no cover
    import importlib.metadata

    PYTEST_61 = importlib.metadata.version("pytest") >= "6.1.0"
except ImportError:  # pragma: no cover
    try:
        import importlib_metadata

        PYTEST_61 = importlib_metadata.version("pytest") >= "6.1.0"
    except ImportError:  # pragma: no cover
        import pkg_resources

        PYTEST_61 = pkg_resources.get_distribution("pytest").version >= "6.1.0"


class FuncWrapper:
    def __init__(self, func):
        self.func = func

    @staticmethod
    def register(func):
        """
        This function will just raise a RuntimeError in case a function
        registration, which also sets a nested namespace, tries to override
        a known helper function with that nested namespace.
        This will just make the raised error make more sense.

        Instead of "AttributeError: 'function' object has no attribute 'register'",
        we will raise the exception below.
        """
        raise RuntimeError(
            "A namespace is already registered under the name: {}".format(func.__name__)
        )

    def __call__(self, *args, **kwargs):
        """
        This wrapper will just call the actual helper function
        """
        __tracebackhide__ = True
        return self.func(*args, **kwargs)


class HelpersRegistry:
    """
    Helper functions registrar which supports namespaces
    """

    __slots__ = ("_registry",)

    def __init__(self):
        self._registry = {}

    def register(self, func, name=None):
        """
        Register's a new function as a helper
        """
        if isinstance(func, str):
            return partial(self.register, name=func)

        if name is None:
            name = func.__name__
        if name in self._registry:
            raise RuntimeError(
                "A helper function is already registered under the name: {}".format(name)
            )
        self._registry[name] = wraps(func)(FuncWrapper(func))
        return func

    def __getattribute__(self, name):
        if name in ("__class__", "_registry", "register"):
            return object.__getattribute__(self, name)
        return self._registry.setdefault(name, self.__class__())

    def __repr__(self):
        return "{} {!r}>".format(self.__class__.__name__, self._registry)

    def __call__(self, *_, **__):
        raise RuntimeError("The helper being called was not registred")

    def __contains__(self, key):
        return key in self._registry

    if PYTEST_61 is False:  # pragma: no cover

        def __fspath__(self):
            # Compatibility with PyTest 6.0.x
            return __file__


def pytest_load_initial_conftests(*_):
    try:
        pytest.helpers  # pragma: no cover
    except AttributeError:
        pytest.helpers = HelpersRegistry()


@pytest.hookimpl(trylast=True)
def pytest_sessionstart(session):
    session.config.pluginmanager.register(pytest.helpers, "helpers-namespace")


def pytest_unconfigure():  # pragma: no cover
    try:
        delattr(pytest, "helpers")
    except AttributeError:
        pass
